from __future__ import annotations

from subprocess import PIPE
from typing import Literal

import click
from delfino import run
from delfino.decorators import pass_args
from delfino.execution import OnError
from delfino.models import AppContext

from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.utils import assert_executable_installed
from delfino_core.vcs_tools import (
    consume_args_until_next_option,
    get_new_branch_name_or_switch_to_branch,
    get_trunk_branch,
    get_vcs_cli_tool,
    title_and_branch_prefix_from_issue_tracker,
)


@click.group("gh")
def run_gh():
    """Extends `gh` or passes through.

    If a `gh` sub-command is not extended, it will be passed through to the original `gh` command.

    See the help of the sub-commands for more details. See also https://cli.github.com/manual/ or `gh --help`.
    """
    assert_executable_installed("git", required_by="gh")
    assert_executable_installed("gh", required_by="gh")


@run_gh.group("pr")
def run_gh_pr():
    """Extends `gh pr` with additional functions and defaults.

    See the help of the sub-commands for more details.
    """


@run_gh_pr.command("create")
@pass_args
def run_gh_pr_create(passed_args: tuple[str, ...]):
    """Extends `gh pr create` with additional functions and defaults.

    Defaults to `--assignee @me --base main --draft` and adds the title if provided.
    Title doesn't need to be quoted. Any word after the command that doesn't start with
    `-` is assumed to be the title.
    """
    title = None
    args = list(passed_args)
    if args and not args[0].startswith("-"):
        # First arg is not an option, assume it's the PR title
        title, args = consume_args_until_next_option(args)

    run(
        [
            "gh",
            "pr",
            "create",
            "--assignee",
            "@me",
            "--base",
            get_trunk_branch(),
            "--draft",
            "--body",
            "",
            *(["--title", title] if title else []),
            *args,
        ],
        on_error=OnError.ABORT,
    )


@run_gh_pr.command("start")
@pass_args
@pass_plugin_app_context
@click.pass_context
def run_gh_pr_start(
    click_context: click.Context,
    app_context: AppContext[CorePluginConfig],
    passed_args: tuple[str, ...],
):
    """Like `gh pr create`, but with additional functionality.

    - Asks for a title if not provided.
    - Stashes any uncommitted changes.
    - Switches to `main` first and pulls the latest changes.
    - Creates a new branch using git/system username and the title.
    - Creates an empty commit to be able to create a PR.
    - Push the branch and the commit to remote.
    - Pops the stash.
    """
    _run_vcs_start(app_context, click_context, passed_args, "gh")


@run_gh_pr.command("view")
@pass_args
def run_gh_pr_view(passed_args: tuple[str, ...]):
    """Extends `gh pr view` with additional functions and defaults.

    Defaults to `--web`.
    """
    run(["gh", "pr", "view", "--web", *passed_args], on_error=OnError.ABORT)


@click.group("glab")
def run_glab():
    """Extends `glab` or passes through.

    If a `glab` sub-command is not extended, it will be passed through to the original `glab` command.

    See the help of the sub-commands for more details. See also https://gitlab.com/gitlab-org/cli or `glab --help`.
    """
    assert_executable_installed("git", required_by="gh")
    assert_executable_installed("glab", required_by="gh")


@run_glab.group("mr")
def run_glab_mr():
    """Extends `glab mr` with additional functions and defaults.

    See the help of the sub-commands for more details.
    """


@run_glab_mr.command("create")
@pass_args
def run_glab_mr_create(passed_args: tuple[str, ...]):
    """Extends `glab mr create` with additional functions and defaults.

    Defaults to `--assignee @me --create-source-branch --target-branch <TRUNK> --draft` and adds the title if provided.
    Title doesn't need to be quoted. Any word after the command that doesn't start with
    `-` is assumed to be the title.
    """
    title = None
    args = list(passed_args)
    if args and not args[0].startswith("-"):
        # First arg is not an option, assume it's the PR title
        title, args = consume_args_until_next_option(args)

    run(
        [
            "glab",
            "mr",
            "create",
            "--assignee",
            "@me",
            "--target-branch",
            get_trunk_branch(),
            "--draft",
            *(["--title", title] if title else []),
            "--web",
            *args,
        ],
        on_error=OnError.ABORT,
    )


@run_glab_mr.command("view")
@pass_args
def run_glab_mr_view(passed_args: tuple[str, ...]):
    """Extends `glab mr view` with additional functions and defaults.

    Defaults to `--web`.
    """
    run(["gh", "mr", "view", "--web", *passed_args], on_error=OnError.ABORT)


@run_glab_mr.command("start")
@pass_args
@pass_plugin_app_context
@click.pass_context
def run_glab_mr_start(
    click_context: click.Context,
    app_context: AppContext[CorePluginConfig],
    passed_args: tuple[str, ...],
):
    """Like `glab mr create`, but with additional functionality.

    - Asks for a title if not provided.
    - Stashes any uncommitted changes.
    - Switches to `main` first and pulls the latest changes.
    - Creates a new branch using git/system username and the title.
    - Creates an empty commit to be able to create a PR.
    - Push the branch and the commit to remote.
    - Pops the stash.
    """
    _run_vcs_start(app_context, click_context, passed_args, "glab")


@click.group("vcs")
def run_vcs():
    """Alias for `gh`/`glab` with auto-detection.

    See the `gh` and `glab` commands for help.
    """
    assert_executable_installed("git", required_by="vcs")

    vcs_cli_tool = get_vcs_cli_tool()
    assert_executable_installed(vcs_cli_tool, required_by="vcs")


@run_vcs.group("mr")
def run_vcs_mr():
    """Alias for `gitlab mr` or `gh pr`, selecting the appropriate one based on the git remote."""


@run_vcs.group("pr")
def run_vcs_pr():
    """Alias for `gitlab mr` or `gh pr`, selecting the appropriate one based on the git remote."""


@run_vcs_pr.command("start")
@pass_args
@pass_plugin_app_context
@click.pass_context
def run_vcs_pr_start(
    click_context: click.Context,
    app_context: AppContext[CorePluginConfig],
    passed_args: tuple[str, ...],
):
    """Alias for `gh pr start`."""
    _run_vcs_start(app_context, click_context, passed_args)


@run_vcs_mr.command("start")
@pass_args
@pass_plugin_app_context
@click.pass_context
def run_vcs_mr_start(
    click_context: click.Context,
    app_context: AppContext[CorePluginConfig],
    passed_args: tuple[str, ...],
):
    """Alias for `glab mr start`."""
    _run_vcs_start(app_context, click_context, passed_args)


def _run_vcs_start(
    app_context: AppContext[CorePluginConfig],
    click_context: click.Context,
    passed_args: tuple[str, ...],
    vcs_cli_tool: Literal["gh", "glab"] | None = None,
):
    args = list(passed_args)
    title, args = consume_args_until_next_option(args)
    if not vcs_cli_tool:
        vcs_cli_tool = get_vcs_cli_tool()

    title, branch_prefix = title_and_branch_prefix_from_issue_tracker(
        vcs_cli_tool, title, app_context.plugin_config.vcs
    )
    branch_name, create_branch = get_new_branch_name_or_switch_to_branch(branch_prefix, title)

    if create_branch:
        run(["git", "stash"], on_error=OnError.ABORT)
        run(["git", "checkout", get_trunk_branch()], on_error=OnError.ABORT)
        run(["git", "pull"], on_error=OnError.ABORT)
        run(["git", "checkout", "-b", branch_name], on_error=OnError.ABORT)

        # Check if last commit messages is not title
        last_commit_message = run(
            ["git", "log", "-1", "--pretty=%B"], stdout=PIPE, on_error=OnError.ABORT
        ).stdout.decode()
        if not last_commit_message.startswith(title):
            # Create and empty commit to be able to create a PR
            run(["git", "commit", "--allow-empty", "-m", title], on_error=OnError.ABORT)

        run(["git", "push", "--set-upstream", "origin", branch_name], on_error=OnError.ABORT)

    # Create the PR using gh
    func = run_gh_pr_create if vcs_cli_tool == "gh" else run_glab_mr_create
    click_context.forward(func, passed_args=["--title", title, *args])

    run(["git", "stash", "pop"], on_error=OnError.ABORT)



import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple

from delfino.decorators import pass_app_context
from delfino.models.pyproject_toml import PluginConfig
from pydantic import BaseModel, Field
from typing_extensions import Annotated

_LOG = logging.getLogger(__name__)


class MypyConfig(BaseModel):
    strict_directories: List[Path] = []


class IssueTrackingConfig(BaseModel):
    _DEFAULT_API_KEY_ENV_VAR = "DELFINO_CORE_ISSUE_TRACKING_API_KEY"
    _DEFAULT_USERNAME_ENV_VAR = "DELFINO_CORE_ISSUE_TRACKING_USERNAME"

    issue_prefix: str = Field(
        "",
        description="Prefix for issue numbers, including a trailing hyphen if used. "
        "If not set, just the issue numbers will be used.",
    )
    tracker_url: str = Field(
        "",
        description="URL for the issue tracker. If not set, issue tracker integration will be disabled."
        "Implemented trackers: Jira.",
    )
    username_env_var: str = Field(
        _DEFAULT_USERNAME_ENV_VAR,
        description=f"Environment variable name for the issue tracking username. "
        f"If not set, '{_DEFAULT_USERNAME_ENV_VAR}' will be used by default.",
    )
    api_key_env_var: str = Field(
        _DEFAULT_API_KEY_ENV_VAR,
        description=f"Environment variable name for the issue tracking API key. "
        f"If not set, '{_DEFAULT_API_KEY_ENV_VAR}' will be used by default.",
    )

    @staticmethod
    def _get_env_var(name: str, purpose: str) -> Optional[str]:
        if (value := os.getenv(name)) is None:
            _LOG.warning(f"{purpose} environment variable '{name}' is not set.")

        return value

    @property
    def api_key(self) -> Optional[str]:
        return self._get_env_var(self.api_key_env_var, "Issue tracking API key")

    @property
    def username(self) -> Optional[str]:
        return self._get_env_var(self.username_env_var, "Issue tracking username")


class VCSConfig(BaseModel):
    branch_prefix: Optional[str] = Field(
        None, description="Prefix for branch names. If not set, git username will be used."
    )
    issue_tracking: Annotated[IssueTrackingConfig, Field(default_factory=IssueTrackingConfig)]


class CorePluginConfig(PluginConfig):
    sources_directory: Path = Path("src")
    tests_directory: Path = Path("tests")
    pytest_modules: List[str] = Field(default_factory=list)
    reports_directory: Path = Path("reports")
    test_types: List[str] = ["unit", "integration"]
    verify_commands: Tuple[str, ...] = ("format", "lint", "mypy", "test")
    lint_commands: Tuple[str, ...] = ("ruff", "pylint", "pycodestyle", "pydocstyle")
    format_commands: Tuple[str, ...] = ("ensure-pre-commit", "pyupgrade", "isort", "black")
    test_commands: Tuple[str, ...] = ("pytest", "coverage-report")
    disable_pre_commit: bool = False
    mypy: Annotated[MypyConfig, Field(default_factory=MypyConfig)]
    vcs: Annotated[VCSConfig, Field(default_factory=VCSConfig)]


pass_plugin_app_context = pass_app_context(plugin_config_type=CorePluginConfig)



"""Linting checks on source code."""

import logging
from functools import lru_cache
from itertools import groupby
from os import getenv
from pathlib import Path
from subprocess import PIPE
from typing import List, Optional, Tuple

import click
from delfino.decorators import files_folders_option, pass_args
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.validation import assert_pip_package_installed, pip_package_installed

from delfino_core.backports import path_is_relative_to
from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.spinner import Spinner
from delfino_core.utils import commands_group_help, execute_commands_group


@click.command("pydocstyle")
@pass_args
@files_folders_option
@pass_plugin_app_context
def run_pydocstyle(
    app_context: AppContext[CorePluginConfig],
    passed_args: Tuple[str, ...],
    files_folders: Tuple[str],
):
    """Run docstring linting on source code.

    Docstring linting is done via pydocstyle. The pydocstyle config can be found in the
    `pyproject.toml` file under `[tool.pydocstyle]`. This ensures compliance with PEP 257,
    with a few exceptions. Note that pylint also carries out additional documentation
    style checks.
    """
    assert_pip_package_installed("pydocstyle")

    spinner = Spinner("pydocstyle", "checking documentation style")
    dirs = build_target_paths(app_context, files_folders, False, False)

    results = run(
        ["pydocstyle", *passed_args, *dirs],
        on_error=OnError.PASS,
        env_update_path={"PYTHONPATH": app_context.plugin_config.sources_directory},
        stdout=PIPE,
        stderr=PIPE,
        running_hook=spinner,
    )

    spinner.print_results(results)


@click.command("ruff")
@pass_args
@files_folders_option
@pass_plugin_app_context
def run_ruff(app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str]):
    """Run ruff."""
    assert_pip_package_installed("ruff")

    dirs = build_target_paths(app_context, files_folders)
    spinner = Spinner("ruff", "checking code")

    action = "check" if passed_args and passed_args[0] == "check" else "format"

    results = run(
        ["ruff", action, *passed_args, *dirs], stdout=PIPE, stderr=PIPE, on_error=OnError.PASS, running_hook=spinner
    )
    spinner.print_results(results)


@click.command("pycodestyle")
@pass_args
@files_folders_option
@pass_plugin_app_context
def run_pycodestyle(
    app_context: AppContext[CorePluginConfig],
    passed_args: Tuple[str, ...],
    files_folders: Tuple[str],
):
    """Run PEP8 checking on code.

    PEP8 checking is done via pycodestyle.

    Why pycodestyle and pylint? So far, pylint does not check against every convention in PEP8. As pylint's
    functionality grows, we should move all PEP8 checking to pylint and remove pycodestyle.
    """
    assert_pip_package_installed("pycodestyle")

    dirs = build_target_paths(app_context, files_folders)

    # TODO(Radek): Implement unofficial config support in pyproject.toml by parsing it
    #  and outputting the result into a supported format?
    #  See:
    #    - https://github.com/PyCQA/pycodestyle/issues/813
    #    - https://github.com/PyCQA/pydocstyle/issues/447
    args = [
        "pycodestyle",
        "--ignore",
        "E501,W503,E231,E203,E402",
        # Ignores explained:
        # - E501: Line length is checked by PyLint
        # - W503: Disable checking of "Line break before binary operator". PEP8 recently (~2019) switched to
        #         "line break before the operator" style, so we should permit this usage.
        # - E231: "missing whitespace after ','" is a false positive. Handled by black formatter.
        "--exclude",
        ".svn,CVS,.bzr,.hg,.git,__pycache__,.tox,*_config_parser.py",
        *passed_args,
        *dirs,
    ]
    spinner = Spinner("pycodestyle", "checking code style")
    result = run(
        args,
        stdout=PIPE,
        stderr=PIPE,
        on_error=OnError.PASS,
        env_update_path={"PYTHONPATH": app_context.plugin_config.sources_directory},
        running_hook=spinner,
    )
    spinner.print_results(result)


@lru_cache(maxsize=1)
def cpu_count():
    if getenv("CI", ""):
        if (cpu_shares := Path("/sys/fs/cgroup/cpu/cpu.shares")).is_file():
            return int(cpu_shares.read_text(encoding="utf-8").strip()) // 1024

    log = logging.getLogger("cpu_count")
    fallback_msg = "Number of CPUs could not be determined. Falling back to 1."

    if pip_package_installed("psutil"):
        import psutil  # pylint: disable=import-outside-toplevel

        if not (count := psutil.cpu_count(logical=False)):
            log.warning(fallback_msg)
            return 1
        return count

    log.warning(f"`psutil` is not installed. {fallback_msg}")
    return 1


def build_target_paths(
    app_context: AppContext[CorePluginConfig],
    files_folders: Optional[Tuple[str]] = None,
    include_tests: bool = True,
    include_commands: bool = True,
) -> List[Path]:
    if files_folders:
        return [Path(path) for path in files_folders]
    plugin_config = app_context.plugin_config
    target_paths: List[Path] = [plugin_config.sources_directory]

    if include_tests and plugin_config.tests_directory.exists():
        target_paths.append(plugin_config.tests_directory)
    if include_commands:
        target_paths.extend(
            folder for folder in app_context.pyproject_toml.tool.delfino.local_command_folders if folder.exists()
        )
    return target_paths


@click.command("pylint")
@files_folders_option
@pass_args
@pass_plugin_app_context
def run_pylint(
    app_context: AppContext[CorePluginConfig],
    passed_args: Tuple[str, ...],
    files_folders: Tuple[str],
):
    """Run pylint on code.

    The bulk of our code conventions are enforced via pylint. The pylint config can be
    found in the `.pylintrc` file.
    """
    assert_pip_package_installed("pylint")

    plugin_config = app_context.plugin_config

    def get_pylintrc_folder(path: Path) -> Path:
        if plugin_config.tests_directory.exists() and path_is_relative_to(path, plugin_config.tests_directory):
            return plugin_config.tests_directory
        return app_context.project_root

    target_paths = build_target_paths(app_context, files_folders)
    grouped_paths = groupby(target_paths, get_pylintrc_folder)
    for pylintrc_folder, paths in grouped_paths:
        checked_dirs = list(paths)
        spinner = Spinner("pylint", f"checking '{', '.join(map(str, checked_dirs))}'")
        results = run(
            [
                "pylint",
                "-j",
                str(cpu_count()),
                "--rcfile",
                pylintrc_folder / ".pylintrc",
                *passed_args,
                *checked_dirs,
            ],
            stdout=PIPE,
            stderr=PIPE,
            on_error=OnError.PASS,  # spinner aborts on error
            env_update_path={"PYTHONPATH": app_context.plugin_config.sources_directory},
            running_hook=spinner,
        )
        spinner.print_results(results)


@click.command("lint", help=commands_group_help("lint"))
@files_folders_option
@pass_plugin_app_context
@click.pass_context
def run_group_lint(
    click_context: click.Context,
    app_context: AppContext[CorePluginConfig],
    **kwargs,
):
    del kwargs  # passed to downstream commands
    execute_commands_group(click_context, app_context.plugin_config)
